@@ -14,26 +14,25 @@
  * limitations under the License.
  */
 
+#include "mongoc-prelude.h"
+
 #ifndef MONGOC_STREAM_TLS_SECURE_TRANSPORT_H
 #define MONGOC_STREAM_TLS_SECURE_TRANSPORT_H
 
-#if !defined (MONGOC_INSIDE) && !defined (MONGOC_COMPILATION)
-# error "Only <mongoc.h> can be included directly."
-#endif
-
 #ifdef MONGOC_ENABLE_SSL_SECURE_TRANSPORT
-#include <bson.h>
+#include <bson/bson.h>
+
+#include "mongoc-macros.h"
 
 BSON_BEGIN_DECLS
 
-mongoc_stream_t *
-mongoc_stream_tls_secure_transport_new (mongoc_stream_t  *base_stream,
-                                        const char       *host,
+MONGOC_EXPORT (mongoc_stream_t *)
+mongoc_stream_tls_secure_transport_new (mongoc_stream_t *base_stream,
+                                        const char *host,
                                         mongoc_ssl_opt_t *opt,
-                                        int               client);
+                                        int client);
 
 BSON_END_DECLS
 
 #endif /* MONGOC_ENABLE_SSL_SECURE_TRANSPORT */
 #endif /* MONGOC_STREAM_TLS_SECURE_TRANSPORT_H */
-
